
-------------------------------------------------------------------------------
-- CONFIGURATION
-------------------------------------------------------------------------------
local size = tonumber(arg[1]) or 2
local boat_capacity = tonumber(arg[2]) or 3
local gamma = 0.6
local iterations = 100

start_state =
{
	left =
	{
		M = size,
		C = size
	},
	right =
	{
		M = 0,
		C = 0
	},
	boat = "left"
}



function is_goal_state(state)
	return state.left.M == 0 and state.left.C == 0
end

function is_fail_state(state)
	local bad_left_side  = state.left.M > 0 and state.left.M < state.left.C
	local bad_right_side = state.right.M > 0 and state.right.M < state.right.C

	return bad_left_side or bad_right_side
end

-------------------------------------------------------------------------------
-- END CONFIGURATION
-------------------------------------------------------------------------------





-------------------------------------------------------------------------------
-- HELPERS
-------------------------------------------------------------------------------
function table.copy(orig)
	local orig_type = type(orig)
    local copy
    if orig_type == 'table' then
        copy = {}
        for orig_key, orig_value in next, orig, nil do
            copy[table.copy(orig_key)] = table.copy(orig_value)
        end
        setmetatable(copy, table.copy(getmetatable(orig)))
    else -- number, string, boolean, etc
        copy = orig
    end
    return copy
end


function table.merge(t1, t2)
   for k,v in ipairs(t2) do
      table.insert(t1, v)
   end

   return t1
end

function table.contains(table, val)
   if type(table) ~= 'table' then return false end
   for i=1,#table do
      if table[i] == val then
         return true
      end
   end
   return false
end

function print_table(tab, padding)
	local padding = padding or ""
	for k,v in pairs(tab) do
		if type(v) == 'table' then
			print(padding..tostring(k).." {}")
			print_table(v, padding.."  ")
		else
			print(padding..tostring(k).."  "..tostring(v))
		end
	end
end

function state_to_string(state)
	local str = ""

	for m = 0, state.left.M do
		if m > 0 then
			str = str.."M"
		end
	end

	for k = 0, state.left.C do
		if k > 0 then
			str = str.."C"
		end
	end

	if state.boat == "left" then
		str = str.."<"
	end

	str = str.."---"

	if state.boat == "right" then
		str = str..">"
	end

	for m = 0, state.right.M do
		if m > 0 then
			str = str.."M"
		end
	end

	for k = 0, state.right.C do
		if k > 0 then
			str = str.."C"
		end
	end

	return str
end


function states_are_equal(state1, state2)
	return  state1.left.M == state2.left.M
		and state1.left.C == state2.left.C
		and state1.boat   == state2.boat
end
-------------------------------------------------------------------------------
-- END HELPERS
-------------------------------------------------------------------------------




function calculate_next_states(state, capacity)

	-- insert new states into the old states table making sure there are no duplicates
	function merge_possible_states(old_states, new_states)
		for _, new_state in pairs(new_states) do
			local insert = true
			for _, old_state in pairs(old_states) do
				if states_are_equal(new_state, old_state) then
					insert = false
					break
				end
			end

			if insert then
				table.insert(old_states, new_state)
			end
		end
	end

	if capacity <= 0 then
		if state.boat == "left" then
			state.boat = "right"
		else
			state.boat = "left"
		end

		if is_fail_state(state) then
			return {}
		end

		return {state}
	end

	local possible_states = {}

	if state.boat == "left" then
		if state.left.M > 0 then
			local new_state = table.copy(state)
			new_state.left.M  = state.left.M - 1
			new_state.right.M = state.right.M + 1

			local substates = calculate_next_states(new_state, capacity - 1)
			merge_possible_states(possible_states, substates)
		end

		if state.left.C > 0 then
			local new_state = table.copy(state)
			new_state.left.C  = state.left.C - 1
			new_state.right.C = state.right.C + 1

			local substates = calculate_next_states(new_state, capacity - 1)
			merge_possible_states(possible_states, substates)
		end
	else
		if state.right.M > 0 then
			local new_state = table.copy(state)
			new_state.right.M = state.right.M - 1
			new_state.left.M  = state.left.M + 1

			local substates = calculate_next_states(new_state, capacity - 1)
			merge_possible_states(possible_states, substates)
		end

		if state.right.C > 0 then
			local new_state = table.copy(state)
			new_state.right.C = state.right.C - 1
			new_state.left.C  = state.left.C + 1

			local substates = calculate_next_states(new_state, capacity - 1)
			merge_possible_states(possible_states, substates)
		end
	end

	return possible_states
end

function get_all_next_states(state, max_capacity)
	if max_capacity < 1 then
		return {}
	end

	local result = {}

	for capacity = 1, max_capacity do
		table.merge(result, calculate_next_states(state, capacity))
	end

	return result
end




function add_new_state_to_graph(new_state, Graph, max_capacity)
	local state_string = state_to_string(new_state)
	if Graph[state_string] then return Graph end

	Graph[state_string] = {}
	Graph[state_string].state = new_state
	Graph[state_string].next_states = get_all_next_states(new_state, max_capacity)

	local q_values = {}
	local r_values = {}
	for k,v in pairs(Graph[state_string].next_states) do
		if is_goal_state(v) then
			r_values[k] = 1
		else
			if states_are_equal(start_state, v) then
				r_values[k] = -1
			else
				r_values[k] = 0
			end
		end
		q_values[k] = 0
	end

	Graph[state_string].r_values = r_values
	Graph[state_string].q_values = q_values

	return Graph
end


function print_graph(graph)
	local graph_length = 0
	for kv in pairs(graph) do
		graph_length = graph_length + 1
	end

	print("\n\n\n Graph: ", graph_length)
	for k,v in pairs(graph) do
		print("")
		print(k)
		print("        R")
		print_table(v.r_values, "        ")
		print("")
		print("        Q")
		for k1,v1 in pairs(v.q_values) do
			print("      ",k1, v1, state_to_string(v.next_states[k1]))
		end
	end
end


function start_learning(graph)
	local current_state_string = state_to_string(start_state)

	-- for every learning iteration
	for i = 1, iterations do

		repeat
			local next_state_index = math.random(#(graph[current_state_string].next_states))
			local next_state = graph[current_state_string].next_states[next_state_index]
			local next_state_string = state_to_string(next_state)

			if graph[next_state_string] == nil then
				add_new_state_to_graph(next_state, graph, boat_capacity)
			end

			local r_value = graph[current_state_string].r_values[next_state_index]
			local max_next_state_q_value = math.max(unpack(graph[next_state_string].q_values))
			local new_q_value = r_value + gamma * max_next_state_q_value
			graph[current_state_string].q_values[next_state_index] = new_q_value

		current_state_string = next_state_string

		until(is_goal_state(next_state))
	end

	return graph
end


function start_playing(graph)
	local current_state_string = state_to_string(start_state)

	print("\n\nUsing Q function...")
	print(current_state_string)

	repeat
		local best_next_state_q_value = -10
		local next_state_candidates = {}  --  contains all neighboring states with best Q values
		for k,v in pairs(graph[current_state_string].q_values) do
			local next_state_string = state_to_string(graph[current_state_string].next_states[k])
			local this_is_explored_node = graph[next_state_string] ~= nil

			if this_is_explored_node then
				if v == best_next_state_q_value then
					table.insert(next_state_candidates, next_state_string)
				end

				if v > best_next_state_q_value then
					next_state_candidates = {}
					table.insert(next_state_candidates, next_state_string)
					best_next_state_q_value = v
				end
			end
		end

		current_state_string = next_state_candidates[math.random(#next_state_candidates)]

		print(current_state_string)

	until(is_goal_state(graph[current_state_string].state))
end


local Graph = {}
add_new_state_to_graph(start_state, Graph, boat_capacity)



local seed = os.time()
--local seed = 1560245279
math.randomseed(seed)
print("seed = ", seed)

print("\n\n")

start_learning(Graph)

print_graph(Graph)

start_playing(Graph)
